import asyncio
import os
from typing import Optional

from openinference.instrumentation.langchain import (
    LangChainInstrumentor,
    get_current_span,
)
from uipath._cli._runtime._contracts import (
    UiPathRuntimeFactory,
)
from uipath._cli.middlewares import MiddlewareResult

from .._tracing import LangChainExporter, _instrument_traceable_attributes
from ._runtime._exception import LangGraphRuntimeError
from ._runtime._runtime import (  # type: ignore[attr-defined]
    LangGraphRuntimeContext,
    LangGraphScriptRuntime,
)
from ._utils._graph import LangGraphConfig


def langgraph_run_middleware(
    entrypoint: Optional[str], input: Optional[str], resume: bool, **kwargs
) -> MiddlewareResult:
    """Middleware to handle LangGraph execution"""
    config = LangGraphConfig()
    if not config.exists:
        return MiddlewareResult(
            should_continue=True
        )  # Continue with normal flow if no langgraph.json

    try:
        context = LangGraphRuntimeContext.with_defaults(**kwargs)
        context.entrypoint = entrypoint
        context.input = input
        context.resume = resume

        _instrument_traceable_attributes()

        def generate_runtime(ctx: LangGraphRuntimeContext) -> LangGraphScriptRuntime:
            runtime = LangGraphScriptRuntime(ctx, ctx.entrypoint)
            # If not resuming and no job id, delete the previous state file
            if not ctx.resume and ctx.job_id is None:
                if os.path.exists(runtime.state_file_path):
                    os.remove(runtime.state_file_path)
            return runtime

        async def execute():
            runtime_factory = UiPathRuntimeFactory(
                LangGraphScriptRuntime,
                LangGraphRuntimeContext,
                runtime_generator=generate_runtime,
            )

            if context.job_id:
                runtime_factory.add_span_exporter(LangChainExporter())

            runtime_factory.add_instrumentor(LangChainInstrumentor, get_current_span)

            await runtime_factory.execute(context)

        asyncio.run(execute())

        return MiddlewareResult(
            should_continue=False,
            error_message=None,
        )

    except LangGraphRuntimeError as e:
        return MiddlewareResult(
            should_continue=False,
            error_message=e.error_info.detail,
            should_include_stacktrace=True,
        )
    except Exception as e:
        return MiddlewareResult(
            should_continue=False,
            error_message=f"Error: {str(e)}",
            should_include_stacktrace=True,
        )



import json
import logging
from typing import Any, Iterable, Optional, Type

import httpx
from jsonschema_pydantic import jsonschema_to_pydantic as create_model  # type: ignore
from langchain_core.caches import BaseCache
from langchain_core.runnables.utils import Output
from langchain_core.tools import BaseTool, StructuredTool
from langgraph.types import interrupt
from pydantic import BaseModel
from uipath import UiPath
from uipath.agent.models.agent import (
    AgentDefinition,
    AgentEscalationChannel,
    AgentEscalationResourceConfig,
    AgentIntegrationToolParameter,
    AgentIntegrationToolResourceConfig,
    AgentProcessToolResourceConfig,
    AgentResourceConfig,
)
from uipath.models import CreateAction, InvokeProcess
from uipath.models.connections import ConnectionTokenType

logger = logging.getLogger(__name__)


def create_process_tool(resource: AgentProcessToolResourceConfig) -> Iterable[BaseTool]:
    async def process(**kwargs) -> BaseModel:
        return interrupt(
            InvokeProcess(
                name=resource.name,
                input_arguments=kwargs,
                process_folder_path=resource.properties.folder_path,
            )
        )

    input_schema = create_model(resource.input_schema)

    class ProcessTool(StructuredTool):
        @property
        def OutputType(self) -> type[Output]:
            return create_model(resource.output_schema)

    yield ProcessTool(
        name=resource.name,
        args_schema=input_schema,
        description=resource.description,
        coroutine=process,
    )


def create_escalation_tool_from_channel(channel: AgentEscalationChannel) -> BaseTool:
    async def escalate(**kwargs) -> BaseModel:
        recipients = channel.recipients
        if len(recipients) > 1:
            logger.warning(
                "Received more than one recipient. Defaulting to first recipient."
            )
        assignee = recipients[0].value if recipients else None
        return interrupt(
            CreateAction(
                title=channel.description,
                data=kwargs,
                assignee=assignee,
                app_name=channel.properties.app_name,
                app_folder_path=None,  # Channels specify folder name but not folder path.
                app_folder_key=channel.properties.resource_key,
                app_key=channel.properties.resource_key,
                app_version=channel.properties.app_version,
            )
        )

    input_schema = create_model(channel.input_schema)

    class EscalationTool(StructuredTool):
        @property
        def OutputType(self) -> type[Output]:
            return create_model(channel.output_schema)

    return EscalationTool(
        name=channel.name,
        args_schema=input_schema,
        description=channel.description,
        coroutine=escalate,
    )


def create_escalation_tool(
    resource: AgentEscalationResourceConfig,
) -> Iterable[BaseTool]:
    for channel in resource.channels:
        yield create_escalation_tool_from_channel(channel)


METHOD_MAP = {"GETBYID": "GET"}


def build_query_params(parameters: list[AgentIntegrationToolParameter]):
    query_params = [
        x for x in parameters if x.field_location == "query" and x.value is not None
    ]
    if query_params:
        return "?" + "&".join(f"{q.name}={q.value}" for q in query_params)
    return ""


def filter_query_params(
    kwargs: dict[str, Any], parameters: list[AgentIntegrationToolParameter]
):
    query_params = {x.name for x in parameters if x.field_location == "query"}
    non_query_params = {x.name for x in parameters if x.field_location != "query"}
    fields_to_ignore = query_params - non_query_params
    return {k: v for k, v in kwargs.items() if k not in fields_to_ignore}


def create_integration_tool(
    resource: AgentIntegrationToolResourceConfig,
) -> Iterable[BaseTool]:
    async def integration(**kwargs) -> BaseModel:
        uipath = UiPath()
        remote_connection = await uipath.connections.retrieve_async(
            resource.properties.connection.id
        )
        token = await uipath.connections.retrieve_token_async(
            resource.properties.connection.id, ConnectionTokenType.BEARER
        )
        tool_url = f"{remote_connection.api_base_uri}/v3/element/instances/{remote_connection.element_instance_id}{resource.properties.tool_path}"
        tool_url = f"{tool_url}{build_query_params(resource.properties.parameters)}"
        tool_url = tool_url.format(**kwargs)

        authorization = f"{token.token_type} {token.access_token}"
        method = METHOD_MAP.get(resource.properties.method, resource.properties.method)
        response = await httpx.AsyncClient().request(
            method,
            tool_url,
            headers={"Authorization": authorization},
            content=json.dumps(
                filter_query_params(kwargs, resource.properties.parameters)
            ),
        )
        return response.json()

    input_schema = create_model(resource.input_schema)

    class IntegrationTool(StructuredTool):
        @property
        def OutputType(self) -> type[Output]:
            return create_model({})

    yield IntegrationTool(
        name=resource.name,
        args_schema=input_schema,
        description=resource.description,
        coroutine=integration,
    )


def create_cached_wrapper_from_tool(
    wrapped: BaseTool, cache: Optional[BaseCache]
) -> BaseTool:
    if cache is None:
        return wrapped
    else:

        async def cached_invocation(**kwargs) -> BaseModel:
            namespace = f"{wrapped.name}.tool_invoke"
            key = str(kwargs)
            cached = cache.lookup(namespace, key)
            if cached:
                return cached[0]
            response = await wrapped.ainvoke(input=kwargs)
            cache.update(namespace, key, [response])
            return response

        input_schema = wrapped.args_schema

        class CachedTool(StructuredTool):
            OutputType: Type[BaseModel] = wrapped.OutputType

        return CachedTool(
            name=wrapped.name,
            args_schema=input_schema,
            description=wrapped.description,
            coroutine=cached_invocation,
        )


def create_cached_wrapper(
    tools: Iterable[BaseTool], cache: Optional[BaseCache]
) -> Iterable[BaseTool]:
    for wrapped in tools:
        yield create_cached_wrapper_from_tool(wrapped, cache)


def create_resource_tool(
    resource: AgentResourceConfig, cache: Optional[BaseCache] = None
) -> Iterable[BaseTool]:
    match resource:
        case AgentProcessToolResourceConfig():
            return create_cached_wrapper(create_process_tool(resource), cache)
        case AgentIntegrationToolResourceConfig():
            return create_cached_wrapper(create_integration_tool(resource), cache)
        case AgentEscalationResourceConfig():
            return create_cached_wrapper(create_escalation_tool(resource), cache)
        case _:
            raise NotImplementedError()


def safe_extract_tools(
    agent_definition: AgentDefinition, cache: Optional[BaseCache] = None
) -> list[BaseTool]:
    tools = []
    for resource in agent_definition.resources:
        try:
            for structured_tool in create_resource_tool(resource, cache):
                tools.append(structured_tool)
        except NotImplementedError:
            logger.warning(f"Unable to convert {resource.name} into a tool.")
    return tools



import uuid
from datetime import datetime
from typing import Any, Dict, List, Optional

from langchain_core.messages import (
    AIMessage,
    AIMessageChunk,
    BaseMessage,
    HumanMessage,
    ToolMessage,
)
from uipath.agent.conversation import (
    UiPathConversationContentPartChunkEvent,
    UiPathConversationContentPartEndEvent,
    UiPathConversationContentPartEvent,
    UiPathConversationContentPartStartEvent,
    UiPathConversationEvent,
    UiPathConversationExchangeEvent,
    UiPathConversationMessage,
    UiPathConversationMessageEndEvent,
    UiPathConversationMessageEvent,
    UiPathConversationMessageStartEvent,
    UiPathConversationToolCallEndEvent,
    UiPathConversationToolCallEvent,
    UiPathConversationToolCallStartEvent,
    UiPathInlineValue,
)


def _new_id() -> str:
    return str(uuid.uuid4())


def _wrap_in_conversation_event(
    msg_event: UiPathConversationMessageEvent,
    exchange_id: Optional[str] = None,
    conversation_id: Optional[str] = None,
) -> UiPathConversationEvent:
    """Helper to wrap a message event into a conversation-level event."""
    return UiPathConversationEvent(
        conversation_id=conversation_id or _new_id(),
        exchange=UiPathConversationExchangeEvent(
            exchange_id=exchange_id or _new_id(),
            message=msg_event,
        ),
    )


def _extract_text(content) -> str:
    """Normalize LangGraph message.content to plain text."""
    if isinstance(content, str):
        return content
    if isinstance(content, list):
        return "".join(
            part.get("text", "")
            for part in content
            if isinstance(part, dict) and part.get("type") == "text"
        )
    return str(content or "")


def uipath_to_human_messages(
    uipath_msg: UiPathConversationMessage,
) -> List[HumanMessage]:
    """
    Converts a UiPathConversationMessage into a list of HumanMessages for LangGraph.
    Supports multimodal content parts (text, external content) and preserves metadata.
    """
    human_messages = []

    # Loop over each content part
    if uipath_msg.content_parts:
        for part in uipath_msg.content_parts:
            data = part.data
            content = ""
            metadata: Dict[str, Any] = {
                "message_id": uipath_msg.message_id,
                "content_part_id": part.content_part_id,
                "mime_type": part.mime_type,
                "created_at": uipath_msg.created_at,
                "updated_at": uipath_msg.updated_at,
            }

            if isinstance(data, UiPathInlineValue):
                content = str(data.inline)

            # Append a HumanMessage for this content part
            human_messages.append(HumanMessage(content=content, metadata=metadata))

    # Handle the case where there are no content parts
    else:
        metadata = {
            "message_id": uipath_msg.message_id,
            "role": uipath_msg.role,
            "created_at": uipath_msg.created_at,
            "updated_at": uipath_msg.updated_at,
        }
        human_messages.append(HumanMessage(content="", metadata=metadata))

    return human_messages


def map_message(
    message: BaseMessage,
    exchange_id: Optional[str] = None,
    conversation_id: Optional[str] = None,
) -> Optional[UiPathConversationEvent]:
    """Convert LangGraph BaseMessage (chunk or full) into a UiPathConversationEvent."""
    message_id = getattr(message, "id", None) or _new_id()
    timestamp = datetime.now().isoformat()

    # --- Streaming AIMessageChunk ---
    if isinstance(message, AIMessageChunk):
        msg_event = UiPathConversationMessageEvent(
            message_id=message.id or _new_id(),
        )

        if message.content == []:
            msg_event.start = UiPathConversationMessageStartEvent(
                role="assistant", timestamp=timestamp
            )
            msg_event.content_part = UiPathConversationContentPartEvent(
                content_part_id=f"chunk-{message.id}-{0}",
                start=UiPathConversationContentPartStartEvent(mime_type="text/plain"),
            )

        elif isinstance(message.content, list) and message.content:
            for chunk in message.content:
                if not isinstance(chunk, dict):
                    continue
                idx = chunk.get("index", 0)
                ctype = chunk.get("type")
                id = chunk.get("id", f"chunk-{message.id}-{idx}")

                # Start of a tool call
                if ctype == "tool_use":
                    msg_event.tool_call = UiPathConversationToolCallEvent(
                        tool_call_id=id,
                        start=UiPathConversationToolCallStartEvent(
                            tool_name=chunk.get("name") or "",
                            arguments=UiPathInlineValue(inline=""),
                            timestamp=timestamp,
                        ),
                    )

                # JSON args streaming (content part for tool args)
                elif ctype == "input_json_delta":
                    text = chunk.get("partial_json", "")
                    # first delta: emit content part start + first chunk
                    if text == "":
                        msg_event.content_part = UiPathConversationContentPartEvent(
                            content_part_id=id,
                            start=UiPathConversationContentPartStartEvent(
                                mime_type="application/json"
                            ),
                        )
                    else:
                        msg_event.content_part = UiPathConversationContentPartEvent(
                            content_part_id=id,
                            chunk=UiPathConversationContentPartChunkEvent(
                                data=text,
                                content_part_sequence=idx,
                            ),
                        )

                # Plain text from assistant
                elif ctype == "text":
                    text = chunk.get("text", "")
                    msg_event.content_part = UiPathConversationContentPartEvent(
                        content_part_id=id,
                        chunk=UiPathConversationContentPartChunkEvent(
                            data=text,
                            content_part_sequence=idx,
                        ),
                    )
        elif isinstance(message.content, str) and message.content:
            msg_event.content_part = UiPathConversationContentPartEvent(
                content_part_id=f"content-{message.id}",
                chunk=UiPathConversationContentPartChunkEvent(
                    data=message.content,
                    content_part_sequence=0,
                ),
            )

        stop_reason = message.response_metadata.get("stop_reason")
        if not message.content and stop_reason in ("tool_use", "end_turn"):
            msg_event.end = UiPathConversationMessageEndEvent(timestamp=timestamp)

        if (
            msg_event.start
            or msg_event.content_part
            or msg_event.tool_call
            or msg_event.end
        ):
            return _wrap_in_conversation_event(msg_event, exchange_id, conversation_id)

        return None

    text_content = _extract_text(message.content)

    # --- HumanMessage ---
    if isinstance(message, HumanMessage):
        return _wrap_in_conversation_event(
            UiPathConversationMessageEvent(
                message_id=message_id,
                start=UiPathConversationMessageStartEvent(
                    role="user", timestamp=timestamp
                ),
                content_part=UiPathConversationContentPartEvent(
                    content_part_id=f"cp-{message_id}",
                    start=UiPathConversationContentPartStartEvent(
                        mime_type="text/plain"
                    ),
                    chunk=UiPathConversationContentPartChunkEvent(data=text_content),
                    end=UiPathConversationContentPartEndEvent(),
                ),
                end=UiPathConversationMessageEndEvent(),
            ),
            exchange_id,
            conversation_id,
        )

    # --- AIMessage ---
    if isinstance(message, AIMessage):
        # Extract first tool call if present
        tool_calls = getattr(message, "tool_calls", []) or []
        first_tc = tool_calls[0] if tool_calls else None

        return _wrap_in_conversation_event(
            UiPathConversationMessageEvent(
                message_id=message_id,
                start=UiPathConversationMessageStartEvent(
                    role="assistant", timestamp=timestamp
                ),
                content_part=(
                    UiPathConversationContentPartEvent(
                        content_part_id=f"cp-{message_id}",
                        start=UiPathConversationContentPartStartEvent(
                            mime_type="text/plain"
                        ),
                        chunk=UiPathConversationContentPartChunkEvent(
                            data=text_content
                        ),
                        end=UiPathConversationContentPartEndEvent(),
                    )
                    if text_content
                    else None
                ),
                tool_call=(
                    UiPathConversationToolCallEvent(
                        tool_call_id=first_tc.get("id") or _new_id(),
                        start=UiPathConversationToolCallStartEvent(
                            tool_name=first_tc.get("name"),
                            arguments=UiPathInlineValue(
                                inline=str(first_tc.get("args", ""))
                            ),
                            timestamp=timestamp,
                        ),
                    )
                    if first_tc
                    else None
                ),
                end=UiPathConversationMessageEndEvent(),
            ),
            exchange_id,
            conversation_id,
        )

    # --- ToolMessage ---
    if isinstance(message, ToolMessage):
        return _wrap_in_conversation_event(
            UiPathConversationMessageEvent(
                message_id=message_id,
                tool_call=UiPathConversationToolCallEvent(
                    tool_call_id=message.tool_call_id,
                    start=UiPathConversationToolCallStartEvent(
                        tool_name=message.name or "",
                        arguments=UiPathInlineValue(inline=""),
                        timestamp=timestamp,
                    ),
                    end=UiPathConversationToolCallEndEvent(
                        timestamp=timestamp,
                        result=UiPathInlineValue(inline=message.content),
                    ),
                ),
            ),
            exchange_id,
            conversation_id,
        )

    # --- Fallback ---
    return _wrap_in_conversation_event(
        UiPathConversationMessageEvent(
            message_id=message_id,
            start=UiPathConversationMessageStartEvent(
                role="assistant", timestamp=timestamp
            ),
            content_part=UiPathConversationContentPartEvent(
                content_part_id=f"cp-{message_id}",
                chunk=UiPathConversationContentPartChunkEvent(data=text_content),
            ),
            end=UiPathConversationMessageEndEvent(),
        ),
        exchange_id,
        conversation_id,
    )



import asyncio
from typing import Optional

from openinference.instrumentation.langchain import (
    LangChainInstrumentor,
    get_current_span,
)
from uipath._cli._dev._terminal import UiPathDevTerminal
from uipath._cli._runtime._contracts import UiPathRuntimeFactory
from uipath._cli._utils._console import ConsoleLogger
from uipath._cli.middlewares import MiddlewareResult

from .._tracing import _instrument_traceable_attributes
from ._runtime._context import LangGraphRuntimeContext
from ._runtime._runtime import LangGraphScriptRuntime

console = ConsoleLogger()


def langgraph_dev_middleware(interface: Optional[str]) -> MiddlewareResult:
    """Middleware to launch the developer terminal"""

    try:
        if interface == "terminal":

            def generate_runtime(
                ctx: LangGraphRuntimeContext,
            ) -> LangGraphScriptRuntime:
                return LangGraphScriptRuntime(ctx, ctx.entrypoint)

            runtime_factory = UiPathRuntimeFactory(
                LangGraphScriptRuntime, LangGraphRuntimeContext, generate_runtime
            )

            _instrument_traceable_attributes()
            runtime_factory.add_instrumentor(LangChainInstrumentor, get_current_span)
            app = UiPathDevTerminal(runtime_factory)
            asyncio.run(app.run_async())
        else:
            console.error(f"Unknown interface: {interface}")
    except KeyboardInterrupt:
        console.info("Debug session interrupted by user")
    except Exception as e:
        console.error(f"Error occurred: {e}")
        return MiddlewareResult(
            should_continue=False,
            should_include_stacktrace=True,
        )

    return MiddlewareResult(should_continue=False)



import copy
import json

from pydantic import TypeAdapter
from uipath.agent.models.agent import AgentDefinition

from tests.tools.conftest import (
    get_file_path,
    uipath_connection_mock,
    uipath_integration_mock,
    uipath_interrupt_mock,
    uipath_token_mock,
)
from uipath_langchain.tools.preconfigured import safe_extract_tools


class TestSafeExtractTools:
    """Test class for tool extraction."""

    async def test_extraction(self, monkeypatch, httpx_mock, mocker) -> None:
        # Arrange
        monkeypatch.setenv("UIPATH_URL", "https://example.com/")
        monkeypatch.setenv("UIPATH_ACCESS_TOKEN", "1234567890")

        agent_definition = get_file_path("agent_definition.json")
        with open(agent_definition) as agent_definition_file:
            agent_definition_json = json.load(agent_definition_file)
            agent_definition = TypeAdapter(AgentDefinition).validate_python(
                agent_definition_json
            )
        tool_response = {"foo": "bar"}

        # Act
        tools = safe_extract_tools(agent_definition)

        # Assert tools and schema definitions
        assert len(tools) == 3
        assert tools[0].input_schema.model_json_schema().get("properties") == {
            "query": {"description": "Query.", "title": "Query", "type": "string"}
        }
        assert tools[0].OutputType.model_json_schema().get("properties") == {}
        assert tools[1].input_schema.model_json_schema().get("properties") == {
            "name": {"description": "Name.", "title": "Name", "type": "string"}
        }
        assert tools[1].OutputType.model_json_schema().get("properties") == {
            "status": {
                "description": "Escalation status.",
                "title": "Status",
                "type": "string",
            }
        }
        assert tools[2].input_schema.model_json_schema().get("properties") == {
            "name": {
                "anyOf": [{"type": "string"}, {"type": "null"}],
                "default": None,
                "title": "Name",
            }
        }
        assert tools[2].OutputType.model_json_schema().get("properties") == {
            "content": {"description": "Output content", "title": "Content"}
        }

        # Invoke Integration
        mock_connection = copy.deepcopy(
            agent_definition_json["resources"][0]["properties"]["connection"]
        )
        mock_connection["element_instance_id"] = 1  # Updated instance id.

        with (
            uipath_connection_mock(httpx_mock, response=mock_connection),
            uipath_token_mock(httpx_mock, response={"accessToken": "accessToken"}),
            uipath_integration_mock(httpx_mock, tool_response),
        ):
            # Act
            response = await tools[0].ainvoke(input={"query": "query"})

            # Assert
            assert response == tool_response

        # Invoke Escalation
        with uipath_interrupt_mock(mocker, tool_response):
            # Act
            response = await tools[1].ainvoke(input={"name": "name"})

            # Assert
            assert response == tool_response

        # Invoke Escalation
        with uipath_interrupt_mock(mocker, tool_response):
            # Act
            response = await tools[2].ainvoke(input={"name": "name"})

            # Assert
            assert response == tool_response



from sqlalchemy import Column, Integer, String, Text, DateTime, ForeignKey
from sqlalchemy.orm import relationship
from database import Base
from datetime import datetime

class Job(Base):
    __tablename__ = "jobs"

    id = Column(String, primary_key=True, index=True)
    agent_job_key = Column(String, index=True, nullable=True)
    agent_job_id = Column(Integer, index=True, nullable=True)
    topic = Column(String, index=True)
    status = Column(String, index=True)  # creating, pending, running, suspended, resumed, faulted, completed
    output = Column(Text)
    error_message = Column(Text, nullable=True)
    human_feedback = Column(Text, nullable=True)
    created_at = Column(DateTime, default=datetime.now)
    completed_at = Column(DateTime, nullable=True)

    messages = relationship("InboxMessage", back_populates="job")

class InboxMessage(Base):
    __tablename__ = "inbox_messages"

    id = Column(String, primary_key=True, index=True)
    job_id = Column(String, ForeignKey("jobs.id"), index=True)
    content = Column(Text)
    status = Column(String, index=True)  # pending, approved, rejected
    human_feedback = Column(Text, nullable=True)
    received_at = Column(DateTime, default=datetime.now)
    approved_at = Column(DateTime, nullable=True)

    job = relationship("Job", back_populates="messages")



from langchain_anthropic import ChatAnthropic
from langchain_community.tools.tavily_search import TavilySearchResults
from langgraph.graph import END, START, MessagesState, StateGraph
from langgraph.prebuilt import create_react_agent
from pydantic import BaseModel

tavily_tool = TavilySearchResults(max_results=5)

llm = ChatAnthropic(model="claude-3-5-sonnet-latest")


research_agent = create_react_agent(
    llm, tools=[tavily_tool], prompt="You are a researcher. DO NOT do any math."
)


class GraphOutput(BaseModel):
    answer: str


async def research_node(state: MessagesState) -> GraphOutput:
    result = await research_agent.ainvoke(state)
    return GraphOutput(answer=result["messages"][-1].content)


# Build the state graph
builder = StateGraph(MessagesState, output=GraphOutput)
builder.add_node("researcher", research_node)

builder.add_edge(START, "researcher")
builder.add_edge("researcher", END)

# Compile the graph
graph = builder.compile()



import logging
import os
from typing import Any, List, Optional, Sequence

from langchain_core.callbacks.base import BaseCallbackHandler
from langchain_core.messages import BaseMessage
from langchain_core.runnables.config import RunnableConfig
from langgraph.checkpoint.sqlite.aio import AsyncSqliteSaver
from langgraph.errors import EmptyInputError, GraphRecursionError, InvalidUpdateError
from uipath._cli._runtime._contracts import (
    UiPathBaseRuntime,
    UiPathErrorCategory,
    UiPathRuntimeResult,
)

from ._context import LangGraphRuntimeContext
from ._conversation import map_message
from ._exception import LangGraphRuntimeError
from ._graph_resolver import AsyncResolver, LangGraphJsonResolver
from ._input import LangGraphInputProcessor
from ._output import LangGraphOutputProcessor

logger = logging.getLogger(__name__)


class LangGraphRuntime(UiPathBaseRuntime):
    """
    A runtime class implementing the async context manager protocol.
    This allows using the class with 'async with' statements.
    """

    def __init__(self, context: LangGraphRuntimeContext, graph_resolver: AsyncResolver):
        super().__init__(context)
        self.context: LangGraphRuntimeContext = context
        self.graph_resolver: AsyncResolver = graph_resolver

    async def execute(self) -> Optional[UiPathRuntimeResult]:
        """
        Execute the graph with the provided input and configuration.

        Returns:
            Dictionary with execution results

        Raises:
            LangGraphRuntimeError: If execution fails
        """

        graph = await self.graph_resolver()
        if not graph:
            return None

        try:
            async with AsyncSqliteSaver.from_conn_string(
                self.state_file_path
            ) as memory:
                self.context.memory = memory

                # Compile the graph with the checkpointer
                compiled_graph = graph.compile(checkpointer=self.context.memory)

                # Process input, handling resume if needed
                input_processor = LangGraphInputProcessor(context=self.context)

                processed_input = await input_processor.process()

                callbacks: List[BaseCallbackHandler] = []

                graph_config: RunnableConfig = {
                    "configurable": {
                        "thread_id": (
                            self.context.execution_id
                            or self.context.job_id
                            or "default"
                        )
                    },
                    "callbacks": callbacks,
                }

                recursion_limit = os.environ.get("LANGCHAIN_RECURSION_LIMIT", None)
                max_concurrency = os.environ.get("LANGCHAIN_MAX_CONCURRENCY", None)

                if recursion_limit is not None:
                    graph_config["recursion_limit"] = int(recursion_limit)
                if max_concurrency is not None:
                    graph_config["max_concurrency"] = int(max_concurrency)

                if self.context.chat_handler or self.is_debug_run():
                    final_chunk: Optional[dict[Any, Any]] = None
                    async for stream_chunk in compiled_graph.astream(
                        processed_input,
                        graph_config,
                        stream_mode=["messages", "updates"],
                        subgraphs=True,
                    ):
                        _, chunk_type, data = stream_chunk
                        if chunk_type == "messages":
                            if self.context.chat_handler:
                                if isinstance(data, tuple):
                                    message, _ = data
                                    event = map_message(
                                        message=message,
                                        conversation_id=self.context.execution_id,
                                        exchange_id=self.context.execution_id,
                                    )
                                    if event:
                                        self.context.chat_handler.on_event(event)
                        elif chunk_type == "updates":
                            if isinstance(data, dict):
                                # data is a dict, e.g. {'agent': {'messages': [...]}}
                                for agent_data in data.values():
                                    if isinstance(agent_data, dict):
                                        messages = agent_data.get("messages", [])
                                        if isinstance(messages, list):
                                            for message in messages:
                                                if isinstance(message, BaseMessage):
                                                    message.pretty_print()
                                final_chunk = data

                    self.context.output = self._extract_graph_result(
                        final_chunk, compiled_graph.output_channels
                    )
                else:
                    # Execute the graph normally at runtime or eval
                    self.context.output = await compiled_graph.ainvoke(
                        processed_input, graph_config
                    )

                # Get the state if available
                try:
                    self.context.state = await compiled_graph.aget_state(graph_config)
                except Exception:
                    pass

                output_processor = await LangGraphOutputProcessor.create(self.context)

                self.context.result = await output_processor.process()

                return self.context.result

        except Exception as e:
            if isinstance(e, LangGraphRuntimeError):
                raise

            detail = f"Error: {str(e)}"

            if isinstance(e, GraphRecursionError):
                raise LangGraphRuntimeError(
                    "GRAPH_RECURSION_ERROR",
                    "Graph recursion limit exceeded",
                    detail,
                    UiPathErrorCategory.USER,
                ) from e

            if isinstance(e, InvalidUpdateError):
                raise LangGraphRuntimeError(
                    "GRAPH_INVALID_UPDATE",
                    str(e),
                    detail,
                    UiPathErrorCategory.USER,
                ) from e

            if isinstance(e, EmptyInputError):
                raise LangGraphRuntimeError(
                    "GRAPH_EMPTY_INPUT",
                    "The input data is empty",
                    detail,
                    UiPathErrorCategory.USER,
                ) from e

            raise LangGraphRuntimeError(
                "EXECUTION_ERROR",
                "Graph execution failed",
                detail,
                UiPathErrorCategory.USER,
            ) from e
        finally:
            pass

    async def validate(self) -> None:
        pass

    async def cleanup(self):
        pass

    def _extract_graph_result(self, final_chunk, output_channels: str | Sequence[str]):
        """
        Extract the result from a LangGraph output chunk according to the graph's output channels.

        Args:
            final_chunk: The final chunk from graph.astream()
            graph: The LangGraph instance

        Returns:
            The extracted result according to the graph's output_channels configuration
        """
        # Unwrap from subgraph tuple format if needed
        if isinstance(final_chunk, tuple) and len(final_chunk) == 2:
            final_chunk = final_chunk[
                1
            ]  # Extract data part from (namespace, data) tuple

        # If the result isn't a dict or graph doesn't define output channels, return as is
        if not isinstance(final_chunk, dict):
            return final_chunk

        # Case 1: Single output channel as string
        if isinstance(output_channels, str):
            if output_channels in final_chunk:
                return final_chunk[output_channels]
            else:
                return final_chunk

        # Case 2: Multiple output channels as sequence
        elif hasattr(output_channels, "__iter__") and not isinstance(
            output_channels, str
        ):
            # Check which channels are present
            available_channels = [ch for ch in output_channels if ch in final_chunk]

            # if no available channels, output may contain the last_node name as key
            unwrapped_final_chunk = {}
            if not available_channels:
                if len(final_chunk) == 1 and isinstance(
                    unwrapped_final_chunk := next(iter(final_chunk.values())), dict
                ):
                    available_channels = [
                        ch for ch in output_channels if ch in unwrapped_final_chunk
                    ]

            if available_channels:
                # Create a dict with the available channels
                return {
                    channel: final_chunk.get(channel, None)
                    or unwrapped_final_chunk[channel]
                    for channel in available_channels
                }

        # Fallback for any other case
        return final_chunk


class LangGraphScriptRuntime(LangGraphRuntime):
    """
    Resolves the graph from langgraph.json config file and passes it to the base runtime.
    """

    def __init__(
        self, context: LangGraphRuntimeContext, entrypoint: Optional[str] = None
    ):
        self.resolver = LangGraphJsonResolver(entrypoint=entrypoint)
        super().__init__(context, self.resolver)

    async def cleanup(self):
        await super().cleanup()
        await self.resolver.cleanup()



"""
Vector store implementation that connects to UiPath Context Grounding as a backend.

This is a read-only vector store that uses the UiPath Context Grounding API to retrieve documents.

You need to set the following environment variables (also see .env.example):
### - UIPATH_URL="https://alpha.uipath.com/{ORG_ID}/{TENANT_ID}"
### - UIPATH_ACCESS_TOKEN={BEARER_TOKEN_WITH_CONTEXT_GROUNDING_PERMISSIONS}
### - UIPATH_FOLDER_PATH="" - this can be left empty
### - UIPATH_FOLDER_KEY="" - this can be left empty
"""

from collections.abc import Iterable
from typing import Any, Optional, TypeVar

from langchain_core.documents import Document
from langchain_core.embeddings import Embeddings
from langchain_core.vectorstores import VectorStore
from uipath import UiPath

VST = TypeVar("VST", bound="ContextGroundingVectorStore")


class ContextGroundingVectorStore(VectorStore):
    """Vector store that uses UiPath Context Grounding (ECS) as a backend.

    This class provides a straightforward implementation that connects to the
    UiPath Context Grounding API for semantic searching.

    Example:
        .. code-block:: python

            from uipath_agents_gym.tools.ecs_vectorstore import ContextGroundingVectorStore

            # Initialize the vector store with an index name
            vectorstore = ContextGroundingVectorStore(index_name="ECCN")

            # Perform similarity search
            docs_with_scores = vectorstore.similarity_search_with_score(
                "How do I process an invoice?", k=5
            )
    """

    def __init__(
        self,
        index_name: str,
        folder_path: Optional[str] = None,
        uipath_sdk: Optional[UiPath] = None,
    ):
        """Initialize the ContextGroundingVectorStore.

        Args:
            index_name: Name of the context grounding index to use
            uipath_sdk: Optional SDK instance to use. If not provided, a new instance will be created.
        """
        self.index_name = index_name
        self.folder_path = folder_path
        self.sdk = uipath_sdk or UiPath()

    def similarity_search_with_score(
        self, query: str, k: int = 4, **kwargs: Any
    ) -> list[tuple[Document, float]]:
        """Return documents most similar to the query along with the distances.
        The distance is 1 - score, where score is the relevance score returned by the Context Grounding API.

        Args:
            query: The query string
            k: Number of results to return (default=4)

        Returns:
            list of tuples of (document, score)
        """
        # Call the UiPath SDK to perform the search
        results = self.sdk.context_grounding.search(
            name=self.index_name,
            query=query,
            number_of_results=k,
            folder_path=self.folder_path,
        )

        # Convert the results to Documents with scores
        docs_with_scores = []
        for result in results:
            # Create metadata from result fields
            metadata = {
                "source": result.source,
                "id": result.id,
                "reference": result.reference,
                "page_number": result.page_number,
                "source_document_id": result.source_document_id,
                "caption": result.caption,
            }

            # Add any operation metadata if available
            if result.metadata:
                metadata["operation_id"] = result.metadata.operation_id
                metadata["strategy"] = result.metadata.strategy

            # Create a Document with the content and metadata
            doc = Document(
                page_content=result.content,
                metadata=metadata,
            )

            score = 1.0 - float(result.score)

            docs_with_scores.append((doc, score))

        return docs_with_scores

    def similarity_search_with_relevance_scores(
        self, query: str, k: int = 4, **kwargs: Any
    ) -> list[tuple[Document, float]]:
        """Return documents along with their relevance scores on a scale from 0 to 1.

        This directly uses the scores provided by the Context Grounding API,
        which are already normalized between 0 and 1.

        Args:
            query: The query string
            k: Number of documents to return (default=4)

        Returns:
            list of tuples of (document, relevance_score)
        """
        return [
            (doc, 1.0 - score)
            for doc, score in self.similarity_search_with_score(query, k, **kwargs)
        ]

    async def asimilarity_search_with_score(
        self, query: str, k: int = 4, **kwargs: Any
    ) -> list[tuple[Document, float]]:
        """Asynchronously return documents most similar to the query along with scores.

        Args:
            query: The query string
            k: Number of results to return (default=4)

        Returns:
            list of tuples of (document, score)
        """
        # Call the UiPath SDK to perform the search asynchronously
        results = await self.sdk.context_grounding.search_async(
            name=self.index_name,
            query=query,
            number_of_results=k,
            folder_path=self.folder_path,
        )

        # Convert the results to Documents with scores
        docs_with_scores = []
        for result in results:
            # Create metadata from result fields
            metadata = {
                "source": result.source,
                "id": result.id,
                "reference": result.reference,
                "page_number": result.page_number,
                "source_document_id": result.source_document_id,
                "caption": result.caption,
            }

            # Add any operation metadata if available
            if result.metadata:
                metadata["operation_id"] = result.metadata.operation_id
                metadata["strategy"] = result.metadata.strategy

            # Create a Document with the content and metadata
            doc = Document(
                page_content=result.content,
                metadata=metadata,
            )

            # Get the distance score as 1 - ecs_score
            score = 1.0 - float(result.score)

            docs_with_scores.append((doc, score))

        return docs_with_scores

    async def asimilarity_search_with_relevance_scores(
        self, query: str, k: int = 4, **kwargs: Any
    ) -> list[tuple[Document, float]]:
        """Asynchronously return documents along with their relevance scores on a scale from 0 to 1.

        This directly uses the scores provided by the Context Grounding API,
        which are already normalized between 0 and 1.

        Args:
            query: The query string
            k: Number of documents to return (default=4)

        Returns:
            list of tuples of (document, relevance_score)
        """
        return [
            (doc, 1.0 - score)
            for doc, score in await self.asimilarity_search_with_score(
                query, k, **kwargs
            )
        ]

    def similarity_search(
        self, query: str, k: int = 4, **kwargs: Any
    ) -> list[Document]:
        """Return documents most similar to the query.

        Args:
            query: The query string
            k: Number of results to return (default=4)

        Returns:
            list of documents most similar to the query
        """
        docs_and_scores = self.similarity_search_with_score(query, k, **kwargs)
        return [doc for doc, _ in docs_and_scores]

    async def asimilarity_search(
        self, query: str, k: int = 4, **kwargs: Any
    ) -> list[Document]:
        """Asynchronously return documents most similar to the query.

        Args:
            query: The query string
            k: Number of results to return (default=4)

        Returns:
            list of documents most similar to the query
        """
        docs_and_scores = await self.asimilarity_search_with_score(query, k, **kwargs)
        return [doc for doc, _ in docs_and_scores]

    @classmethod
    def from_texts(
        cls: type[VST],
        texts: list[str],
        embedding: Embeddings,
        metadatas: Optional[list[dict[str, Any]]] = None,
        **kwargs: Any,
    ) -> VST:
        """This method is required by the VectorStore abstract class, but is not supported
        by ContextGroundingVectorStore which is read-only.

        Raises:
            NotImplementedError: This method is not supported by ContextGroundingVectorStore
        """
        raise NotImplementedError(
            "ContextGroundingVectorStore is a read-only wrapper for UiPath Context Grounding. "
            "Creating a vector store from texts is not supported."
        )

    # Other required methods with minimal implementation to satisfy the interface
    def add_texts(
        self,
        texts: Iterable[str],
        metadatas: Optional[list[dict[str, Any]]] = None,
        **kwargs: Any,
    ) -> list[str]:
        """Not implemented for ContextGroundingVectorStore as this is a read-only wrapper."""
        raise NotImplementedError(
            "ContextGroundingVectorStore is a read-only wrapper for UiPath Context Grounding."
        )

    def delete(self, ids: Optional[list[str]] = None, **kwargs: Any) -> Optional[bool]:
        """Not implemented for ContextGroundingVectorStore as this is a read-only wrapper."""
        raise NotImplementedError(
            "ContextGroundingVectorStore is a read-only wrapper for UiPath Context Grounding."
        )



import os
from contextlib import contextmanager


def get_file_path(file_name: str) -> str:
    """
    Get the path to a mock file, checking multiple possible locations.

    Args:
        file_name: Name of the file to locate

    Returns:
        Absolute path to the file
    """
    if os.path.isfile(f"mocks/{file_name}"):
        return os.path.abspath(f"mocks/{file_name}")
    else:
        return os.path.abspath(f"tests/tools/mocks/{file_name}")


@contextmanager
def uipath_connection_mock(httpx_mock, response):
    httpx_mock.add_response(
        url="https://example.com/api/v1/Connections/connection-id",
        method="GET",
        json=response,
    )

    try:
        yield
    finally:
        pass


@contextmanager
def uipath_token_mock(httpx_mock, response):
    httpx_mock.add_response(
        url="https://example.com/api/v1/Connections/connection-id/token?tokenType=bearer",
        method="GET",
        json=response,
    )

    try:
        yield
    finally:
        pass


@contextmanager
def uipath_integration_mock(httpx_mock, response):
    httpx_mock.add_response(
        url="https://example.uipath.com//v3/element/instances/0/integration",
        method="POST",
        json=response,
    )

    try:
        yield
    finally:
        pass


@contextmanager
def uipath_interrupt_mock(mocker, tool_response):
    mocker.patch(
        "uipath_langchain.tools.preconfigured.interrupt", return_value=tool_response
    )

    try:
        yield
    finally:
        pass



import sys
from contextlib import asynccontextmanager

from langchain_anthropic import ChatAnthropic
from langchain_mcp_adapters.client import MultiServerMCPClient
from langgraph.prebuilt import create_react_agent

model = ChatAnthropic(model="claude-3-5-sonnet-latest")


@asynccontextmanager
async def make_graph():
    client = MultiServerMCPClient({
            "math": {
                "command": sys.executable,
                "args": ["src/simple-local-mcp/math_server.py"],
                "transport": "stdio",
            },
            "weather": {
                "command": sys.executable,
                "args": ["src/simple-local-mcp/weather_server.py"],
                "transport": "stdio",
            },
        })
    agent = create_react_agent(model, await client.get_tools())
    yield agent



import asyncio
from typing import Any, Awaitable, Callable, Optional

from langgraph.graph.state import CompiledStateGraph, StateGraph
from uipath._cli._runtime._contracts import (
    UiPathErrorCategory,
)

from .._utils._graph import GraphConfig, LangGraphConfig
from ._exception import LangGraphRuntimeError


class LangGraphJsonResolver:
    def __init__(self, entrypoint: Optional[str] = None) -> None:
        self.entrypoint = entrypoint
        self.graph_config: Optional[GraphConfig] = None
        self._lock = asyncio.Lock()
        self._graph_cache: Optional[StateGraph[Any, Any, Any]] = None
        self._resolving: bool = False

    async def __call__(self) -> StateGraph[Any, Any, Any]:
        # Fast path: if already resolved, return immediately without locking
        if self._graph_cache is not None:
            return self._graph_cache

        # Slow path: acquire lock and resolve
        async with self._lock:
            # Double-check after acquiring lock (another coroutine may have resolved it)
            if self._graph_cache is not None:
                return self._graph_cache

            self._graph_cache = await self._resolve(self.entrypoint)
            return self._graph_cache

    async def _resolve(self, entrypoint: Optional[str]) -> StateGraph[Any, Any, Any]:
        config = LangGraphConfig()
        if not config.exists:
            raise LangGraphRuntimeError(
                "CONFIG_MISSING",
                "Invalid configuration",
                "Failed to load configuration",
                UiPathErrorCategory.DEPLOYMENT,
            )

        try:
            config.load_config()
        except Exception as e:
            raise LangGraphRuntimeError(
                "CONFIG_INVALID",
                "Invalid configuration",
                f"Failed to load configuration: {str(e)}",
                UiPathErrorCategory.DEPLOYMENT,
            ) from e

        # Determine entrypoint if not provided
        graphs = config.graphs
        if not entrypoint and len(graphs) == 1:
            entrypoint = graphs[0].name
        elif not entrypoint:
            graph_names = ", ".join(g.name for g in graphs)
            raise LangGraphRuntimeError(
                "ENTRYPOINT_MISSING",
                "Entrypoint required",
                f"Multiple graphs available. Please specify one of: {graph_names}.",
                UiPathErrorCategory.DEPLOYMENT,
            )

        # Get the specified graph
        self.graph_config = config.get_graph(entrypoint)
        if not self.graph_config:
            raise LangGraphRuntimeError(
                "GRAPH_NOT_FOUND",
                "Graph not found",
                f"Graph '{entrypoint}' not found.",
                UiPathErrorCategory.DEPLOYMENT,
            )
        try:
            loaded_graph = await self.graph_config.load_graph()
            return (
                loaded_graph.builder
                if isinstance(loaded_graph, CompiledStateGraph)
                else loaded_graph
            )
        except ImportError as e:
            raise LangGraphRuntimeError(
                "GRAPH_IMPORT_ERROR",
                "Graph import failed",
                f"Failed to import graph '{entrypoint}': {str(e)}",
                UiPathErrorCategory.USER,
            ) from e
        except TypeError as e:
            raise LangGraphRuntimeError(
                "GRAPH_TYPE_ERROR",
                "Invalid graph type",
                f"Graph '{entrypoint}' is not a valid StateGraph or CompiledStateGraph: {str(e)}",
                UiPathErrorCategory.USER,
            ) from e
        except ValueError as e:
            raise LangGraphRuntimeError(
                "GRAPH_VALUE_ERROR",
                "Invalid graph value",
                f"Invalid value in graph '{entrypoint}': {str(e)}",
                UiPathErrorCategory.USER,
            ) from e
        except Exception as e:
            raise LangGraphRuntimeError(
                "GRAPH_LOAD_ERROR",
                "Failed to load graph",
                f"Unexpected error loading graph '{entrypoint}': {str(e)}",
                UiPathErrorCategory.USER,
            ) from e

    async def cleanup(self):
        """Clean up resources"""
        async with self._lock:
            if self.graph_config:
                await self.graph_config.cleanup()
                self.graph_config = None
            self._graph_cache = None


AsyncResolver = Callable[[], Awaitable[StateGraph[Any, Any, Any]]]


class LangGraphJsonResolverContext:
    """
    Async context manager wrapping LangGraphJsonResolver.
    Returns a callable that can be passed directly as AsyncResolver to LangGraphRuntime.
    Thread-safe and reuses the same resolved graph across concurrent executions.
    """

    def __init__(self, entrypoint: Optional[str] = None) -> None:
        self._resolver = LangGraphJsonResolver(entrypoint)

    async def __aenter__(self) -> AsyncResolver:
        # Return a callable that safely reuses the cached graph
        async def resolver_callable() -> StateGraph[Any, Any, Any]:
            return await self._resolver()

        return resolver_callable

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        await self._resolver.cleanup()



import json
import logging
from typing import Any, Dict, List, Optional

from opentelemetry.sdk.trace.export import SpanExportResult
from uipath.tracing import LlmOpsHttpExporter

logger = logging.getLogger(__name__)


def _safe_parse_json(s: Any) -> Any:
    """Safely parse a JSON string, returning the original if not a string or on error."""
    if not isinstance(s, str):
        return s
    try:
        return json.loads(s)
    except (json.JSONDecodeError, TypeError):
        return s


def _get_llm_messages(attributes: Dict[str, Any], prefix: str) -> List[Dict[str, Any]]:
    """Extracts and reconstructs LLM messages from flattened attributes."""
    messages: dict[int, dict[str, Any]] = {}
    message_prefix = f"{prefix}."

    for key, value in attributes.items():
        if key.startswith(message_prefix):
            parts = key[len(message_prefix) :].split(".")
            if len(parts) >= 2 and parts[0].isdigit():
                index = int(parts[0])
                if index not in messages:
                    messages[index] = {}
                current: Any = messages[index]

                for i, part in enumerate(parts[1:-1]):
                    key_part: str | int = part
                    if part.isdigit() and (
                        i + 2 < len(parts) and parts[i + 2].isdigit()
                    ):
                        key_part = int(part)

                    if isinstance(current, dict):
                        if key_part not in current:
                            current[key_part] = {}
                        current = current[key_part]
                    elif isinstance(current, list) and isinstance(key_part, int):
                        if key_part >= len(current):
                            current.append({})
                        current = current[key_part]

                current[parts[-1]] = value

    # Convert dict to list, ordered by index
    return [messages[i] for i in sorted(messages.keys())]


class LangChainExporter(LlmOpsHttpExporter):
    # Mapping of old attribute names to new attribute names or (new name, function)
    ATTRIBUTE_MAPPING: dict[str, str | tuple[str, Any]] = {
        "input.value": ("input", _safe_parse_json),
        "output.value": ("output", _safe_parse_json),
        "llm.model_name": "model",
    }

    # Mapping of span types
    SPAN_TYPE_MAPPING: dict[str, str] = {
        "LLM": "completion",
        "TOOL": "toolCall",
        # Add more mappings as needed
    }

    class Status:
        SUCCESS = 1
        ERROR = 2
        INTERRUPTED = 3

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    def _map_llm_call_attributes(self, attributes: Dict[str, Any]) -> Dict[str, Any]:
        """Maps attributes for LLM calls, handling flattened keys."""
        result = attributes.copy()  # Keep original attributes including basic mappings

        # Token Usage
        token_keys = {
            "llm.token_count.prompt": "promptTokens",
            "llm.token_count.completion": "completionTokens",
            "llm.token_count.total": "totalTokens",
        }
        usage = {
            new_key: attributes.get(old_key)
            for old_key, new_key in token_keys.items()
            if old_key in attributes
        }
        if usage:
            result["usage"] = usage

        # Input/Output Messages
        result["input"] = _get_llm_messages(attributes, "llm.input_messages")
        output_messages = _get_llm_messages(attributes, "llm.output_messages")
        result["output"] = output_messages

        # Invocation Parameters
        invocation_params = _safe_parse_json(
            attributes.get("llm.invocation_parameters", "{}")
        )
        if isinstance(invocation_params, dict):
            result["model"] = invocation_params.get("model", result.get("model"))
            settings: dict[str, Any] = {}
            if "max_tokens" in invocation_params:
                settings["maxTokens"] = invocation_params["max_tokens"]
            if "temperature" in invocation_params:
                settings["temperature"] = invocation_params["temperature"]
            if settings:
                result["settings"] = settings

        # Tool Calls
        tool_calls: list[dict[str, Any]] = []
        for msg in output_messages:
            # Ensure msg is a dictionary before proceeding
            if not isinstance(msg, dict):
                continue
            msg_tool_calls = msg.get("message", {}).get("tool_calls", [])

            # Ensure msg_tool_calls is a list
            if not isinstance(msg_tool_calls, list):
                continue

            for tc in msg_tool_calls:
                if not isinstance(tc, dict):
                    continue
                tool_call_data = tc.get("tool_call", {})
                if not isinstance(tool_call_data, dict):
                    continue
                tool_calls.append(
                    {
                        "id": tool_call_data.get("id"),
                        "name": tool_call_data.get("function", {}).get("name"),
                        "arguments": _safe_parse_json(
                            tool_call_data.get("function", {}).get("arguments", "{}")
                        ),
                    }
                )
        if tool_calls:
            result["toolCalls"] = tool_calls

        return result

    def _map_tool_call_attributes(self, attributes: Dict[str, Any]) -> Dict[str, Any]:
        """Maps attributes for tool calls."""
        result = attributes.copy()  # Keep original attributes

        result["type"] = "toolCall"
        result["callId"] = attributes.get("call_id") or attributes.get("id")
        result["toolName"] = attributes.get("tool.name")
        result["arguments"] = _safe_parse_json(
            attributes.get("input", attributes.get("input.value", "{}"))
        )
        result["toolType"] = "Integration"
        result["result"] = _safe_parse_json(
            attributes.get("output", attributes.get("output.value"))
        )
        result["error"] = None

        return result

    def _determine_status(self, error: Optional[str]) -> int:
        if error:
            if error and error.startswith("GraphInterrupt("):
                return self.Status.INTERRUPTED
            return self.Status.ERROR
        return self.Status.SUCCESS

    def _process_span_attributes(self, span_data: Dict[str, Any]) -> Dict[str, Any]:
        """Extracts, transforms, and maps attributes for a span."""
        if "Attributes" not in span_data:
            return span_data

        attributes_val = span_data["Attributes"]
        if isinstance(attributes_val, str):
            try:
                attributes: Dict[str, Any] = json.loads(attributes_val)
            except json.JSONDecodeError as e:
                logger.warning(f"Failed to parse attributes JSON: {e}")
                return span_data
        elif isinstance(attributes_val, dict):
            attributes = attributes_val
        else:
            return span_data

        # Determine SpanType
        if "openinference.span.kind" in attributes:
            span_type = attributes["openinference.span.kind"]
            span_data["SpanType"] = self.SPAN_TYPE_MAPPING.get(span_type, span_type)

        # Apply basic attribute mapping
        for old_key, mapping in self.ATTRIBUTE_MAPPING.items():
            if old_key in attributes:
                if isinstance(mapping, tuple):
                    new_key, func = mapping
                    attributes[new_key] = func(attributes[old_key])
                else:
                    new_key = mapping
                    attributes[new_key] = attributes[old_key]

        # Apply detailed mapping based on SpanType
        span_type = span_data.get("SpanType")
        if span_type == "completion":
            processed_attributes = self._map_llm_call_attributes(attributes)
        elif span_type == "toolCall":
            processed_attributes = self._map_tool_call_attributes(attributes)
        else:
            processed_attributes = attributes.copy()

        span_data["Attributes"] = json.dumps(processed_attributes)

        # Determine status based on error information
        error = attributes.get("error") or attributes.get("exception.message")
        status = self._determine_status(error)
        span_data["Status"] = status

        return span_data

    def _send_with_retries(
        self, url: str, payload: List[Dict[str, Any]], max_retries: int = 4
    ) -> SpanExportResult:
        # Transform attributes in each span's payload before sending
        transformed_payload = [self._process_span_attributes(span) for span in payload]

        return super()._send_with_retries(
            url=url,
            payload=transformed_payload,
            max_retries=max_retries,
        )



from pydantic import BaseModel
from typing import Optional, List
from datetime import datetime

class JobRequest(BaseModel):
    topic: str
    callback_url: Optional[str] = None

class JobResponse(BaseModel):
    job_id: str
    status: str
    message: str

class InboxMessageResponse(BaseModel):
    id: str
    job_id: str
    content: str
    status: str
    human_feedback: Optional[str] = None
    received_at: datetime
    approved_at: Optional[datetime] = None

    class Config:
        orm_mode = True

class MessageApprovalRequest(BaseModel):
    human_feedback: str

class MessageApprovalResponse(BaseModel):
    status: str
    message: str

class JobBasicResponse(BaseModel):
    id: str
    topic: str
    status: str
    created_at: datetime
    completed_at: Optional[datetime] = None

    class Config:
        orm_mode = True

class JobDetailResponse(BaseModel):
    id: str
    topic: str
    status: str
    output: Optional[str] = None
    agent_job_id: Optional[int] = None
    agent_job_key: Optional[str] = None
    human_feedback: Optional[str] = None
    created_at: datetime
    completed_at: Optional[datetime] = None
    error_message: Optional[str] = None
    messages: List[InboxMessageResponse] = []

    class Config:
        orm_mode = True
